/+
 ' Module          : dateutils.d - General Date Utility Functions
 ' Version         : v0.3 beta 
 ' Author          : David L. 'SpottedTiger' Davis
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 22.Jul.04 Compiled and Tested under dmd v0.96.
 ' Modified Date   : 21.Aug.04 Compiled and Tested under dmd v0.99.
 '                 : 11.Nov.04 Recomplied and Tested with dmd v0.106.
 '                 : 20.Jan.05 Recomplied and Tested with dmd v0.111.
 '                 : 20.Mar.05 Recomplied and Tested with dmd v0.119.
 '                 : 10.Jun.05 Recomplied and Tested with dmd v0.126.
 '                 : 08.Jul.05 Added the getDateDiff(), getDateAdd(),
 '                 :           getDateSerial(), and getDatePart() functions, 
 '                 :           plus recomplied and tested with dmd v0.127.
 '                 : 17.Sep.05 Recomplied and Tested with dmd v0.131.
 '                 : 26.Oct.05 Recompiled and Tested with dmd v0.137. 
 '                 : 30.Dec.05 Recompiled and Tested with dmd v0.142.  
 '                 : 22.Mar.06 Recompiled and tested with dmd v0.150. 
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159. 
 '                 : 19.Jul.06 All "Imports" are now private by default 
 '                 :           (which shouldn't be a problem, since I
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Recompiled and Tested with dmd v0.172. 
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 :
 ' Requirements    : std.date, std.string, and std.stdio.
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 '
 ' A good online calendar site at: http://www.timeanddate.com/calendar/index.html 
 '
 ' To complie for a unittest:
 ' dmd dateutils.d -debug=dateutils -unittest
 '
 ' To compile a D program that use dateutils.d:
 ' dmd MySource.d dateutils.d
 '
 ' NOTE: That the Date string parameters can be formatted as
 '       "MM/DD/YYYY", "YYYY/MM/DD"
 '       "DD MMM YYYY", MMM DD YYYY", "YYYY MMM DD",
 '       "D M. YYYY", "M. D YYYY", YYYY M. D".
 '
 ' Public class member functions:
 ' ------------------------------
 ' 01) int calcAge( in char[] )
 ' 02) int calcAge( in d_time )
 ' 03) int getDayOfYear( in char[] = "" ) 
 ' 04) int getDayOfYear( in d_time ) 
 ' 05) int getWeekOfYear( in char[] = "" )
 ' 06) int getWeekOfYear( in d_time )
 ' 07) bool isWeekday( in char[] = "" )
 ' 08) bool isWeekday( in d_time )
 ' 09) bool isWeekend( in char[] = "" )
 ' 10) bool isWeekend( in d_time )
 ' 11) int getLastDayOfMonth( in char[] = "" )
 ' 12) int getLastDayOfMonth( in d_time )
 ' 13) int getLastDayOfMonth( in int, in int )
 ' 14) int getFirstWeekdayOfMonth( in char[] = "" )
 ' 15) int getFirstWeekdayOfMonth( in d_time )
 ' 16) int getLastWeekdayOfMonth( in char[] sDate = "" )
 ' 17) int getLastWeekdayOfMonth( in d_time )
 ' 18) int getQuarterInYear( in char[] = "" )
 ' 19) int getQuarterInYear( in d_time )
 ' 20) d_time parse( in char[], out int, out int, out int )
 ' 21) bool inLeapYear( in char[] = "" )
 ' 22) bool inLeapYear( in d_time = d_time_nan )
 ' 23) long getDateDiff( in char[], in char[], in char[] )
 ' 24) long getDateDiff( in char[], in d_time, in d_time )
 ' 25) d_time getDateAdd( in char[], in long, in d_time )
 ' 26) d_time getDateSerial( in long, in long, in long )
 ' 27) long getDatePart( in char[], in d_time )
 ' 28) long roundToLong( in real )
 ' 29) int roundToInteger( in real )
 ' 30) d_time getDateOnly(in d_time )
 ' 31) d_time getTimeOnly(in d_time )
 ' 32) d_time createNewDateTime( in int, in int, in int, in int, 
 '                               in int, in int, in int )
 +/
module dateutils;

private import std.stdio; // for sprintf() and writefln()
private import std.date;
private import std.math;
private import std.string;

// Define the days of the week
enum : int
{
    SUNDAY    = 0,
    MONDAY    = 1,
    TUESDAY   = 2,
    WEDNESDAY = 3,
    THURSDAY  = 4,
    FRIDAY    = 5,
    SATURDAY  = 6
}

/+
 ' Since my local parse( in char[], out int, out int, out int )
 ' function cancels out std.date's parse( in char[] ), alias it
 ' to bring it into the current namespace.
 +/ 
alias std.date.parse parse;

long roundToLong(in real r)    { return cast(long)((r < 0.0L) ? r - 0.5L : r + 0.5L); }
int  roundToInteger(in real r) { return cast(int)((r < 0.0L) ? r - 0.5L : r + 0.5L);  }

/+
 ' int calcAge( in char[] )
 '
 ' Calcuates age based on the date given compared to the current date.
 '
 ' Returns 0 to int.max
 +/
int calcAge(in char[] sDate)
{
    d_time dtDay   = d_time_nan;
    d_time dtToday = d_time_nan;
    int    iPYear  = 0;
    int    iPWeek  = 0;
    int    iTYear  = 0;
    int    iTWeek  = 0;
    
    dtDay   = std.date.parse(sDate); // LocalTime
    dtToday = std.date.UTCtoLocalTime(std.date.getUTCtime());

    // Pull the Year and Week from the dates for comparing
    iPYear = std.date.YearFromTime(dtDay);
    iPWeek = getWeekOfYear(dtDay);

    iTYear = std.date.YearFromTime(dtToday);
    iTWeek = getWeekOfYear(dtToday);

    // Calculate the differences in years between Today and Birthday
    if ((iPYear != iTYear) && (iPYear < iTYear))
        return ((iTYear - iPYear) - (iTWeek < iPWeek ? 1 : 0));
    else
        return 0;
}

int calcAge(in d_time dtDay)
{
    d_time dtToday = d_time_nan;
    int    iPYear  = 0;
    int    iPWeek  = 0;
    int    iTYear  = 0;
    int    iTWeek  = 0;
    
    dtToday = std.date.UTCtoLocalTime(std.date.getUTCtime());

    // Pull the Year and Week from the dates for comparing
    iPYear = std.date.YearFromTime(dtDay);
    iPWeek = getWeekOfYear(dtDay);

    iTYear = std.date.YearFromTime(dtToday);
    iTWeek = getWeekOfYear(dtToday);
   
    // Calculate the differences in years between Today and Birthday
    if ((iPYear != iTYear) && (iPYear < iTYear))
        return ((iTYear - iPYear) - (iTWeek < iPWeek ? 1 : 0));
    else
        return 0;
} 

/+
 ' int getDayOfYear( in char[] = "" )
 '
 ' Returns the number of days into the year of the date passed in,
 ' and if no date is passed in, then the current date will be used.
 '
 ' Returns 1 to 366 (Days)
 +/
int getDayOfYear(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    int    iYear  = 0;
    int    iDay   = 0;
    int    iMonth = 0;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    iYear  = std.date.YearFromTime(dtDate);
    iMonth = std.date.MonthFromTime(dtDate);
    
    for ( int i = 0; i < iMonth; i++ )
        iDay += getLastDayOfMonth(iYear, i);
        
    //writefln("iDay=%d, iMonth=%d, Date=%d",
    //         iDay, iMonth, std.date.DateFromTime(dtDate));
   
    return iDay + std.date.DateFromTime(dtDate);
} 

int getDayOfYear(in d_time dtDate)
{
    int iYear  = 0;
    int iDay   = 0;
    int iMonth = 0;
   
    iYear  = std.date.YearFromTime(dtDate);
    iMonth = std.date.MonthFromTime(dtDate);
    
    for ( int i = 0; i < iMonth; i++ )
        iDay += getLastDayOfMonth(iYear, i);

    //writefln("iDay=%d, iMonth=%d, Date=%d",
    //         iDay, iMonth, std.date.DateFromTime(dtDate));

    return iDay + std.date.DateFromTime(dtDate);
}

/+
 ' int getWeekOfYear( in char[] = "" )
 '
 ' Returns the week (1 to 53) from the date passed in.
 +/
int getWeekOfYear(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());
        
    return getWeekOfYear(dtDate);
} 

int getWeekOfYear(in d_time dtDate)
{
    d_time dtWDate   = d_time_nan;
    int    iYear     = 0;
    int    iWeek     = 0;
    int    iBypassed = 0;

    iYear = std.date.YearFromTime(dtDate);

    // "01/01/yyyy 00:00:00 UTC" (yyyy = iYear)
    dtWDate = createNewDateTime(iYear, 0, 1, 0, 0, 0, 0);

    //writefln("iYear=%d, dtWDate=\"%s\"", iYear, toUTCString(dtWDate));
    while (std.date.WeekDay(dtWDate) != MONDAY)
    {
        dtWDate += std.date.TicksPerDay;
        iBypassed++;
    }
        
    do 
    {
        if (std.date.WeekDay(dtWDate) == MONDAY)
            iWeek++;
                
        dtWDate += std.date.TicksPerDay;
            
    } while (dtWDate <= dtDate)
       
    //iWeek--;
    if (iBypassed > 1)      
        iWeek++;

    return iWeek;
} 

/+
 ' bool isWeekday( in char[] = "" )
 '
 ' Returns true is the date falls on a weekday.
 +/
bool isWeekday(in char[] sDate = ""c)
{
    d_time dtDate     = d_time_nan;
    int    iDayOfWeek = 0;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());
    
    return isWeekday(dtDate);
} 

bool isWeekday(in d_time dtDate)
{
    int iDayOfWeek = 0;
    
    iDayOfWeek = std.date.WeekDay(dtDate);    
    
    if (iDayOfWeek > 0 && iDayOfWeek < 6)   
        return true;
    else    
        return false;
} 

/+
 ' bool isWeekend( char[] = "" )
 '
 ' Returns true is the date falls on a weekend.
  +/
bool isWeekend(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    return isWeekend(dtDate);
} 

bool isWeekend(in d_time dtDate)
{
    int iDayOfWeek = 0;
    
    iDayOfWeek = std.date.WeekDay(dtDate);    
    
    if (iDayOfWeek == 0 || iDayOfWeek == 6)   
        return true;
    else    
        return false;
}

/+
 ' int getLastDayOfMonth( in char[] = "" )
 '
 ' Returns the max number of days (28, 29, 30, or 31)  
 ' for the month based on the date passed in.
 +/
int getLastDayOfMonth(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());
   
    return getLastDayOfMonth(dtDate); 
} 

int getLastDayOfMonth(in d_time dtDate)
{
    const int[] DPM = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];

    int iDays  = 0;
    int iMonth = 0;
    int iYear  = 0;
    
    // Months return as 0 thru 11
    iMonth = std.date.MonthFromTime(dtDate);
    iDays  = DPM[iMonth];
    
    // Jan = 0, Feb = 1
    if (iMonth == 1)
    {
        iYear = std.date.YearFromTime(dtDate);
        iDays = iDays + std.date.LeapYear(iYear);
    }
            
    return iDays;
} 

int getLastDayOfMonth(in int iYear, in int iMonth)
{
    const int[] DPM = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];

    int iDays = 0;
    int imm   = (iMonth < 0 ? iMonth * -1 : iMonth);
    
    if (imm >= 12)
        imm -= (imm / 12) * 12;

    //writefln("imm=%d", imm);

    iDays = DPM[imm];

    // Jan = 0, Feb = 1...
    if (imm == 1)
        iDays = iDays + std.date.LeapYear(iYear);
            
    return iDays;
} 

/+
 ' int getFirstWeekdayOfMonth( in char[] = "" )
 '
 ' Returns the first weekday in the month of the given date.
 +/
int getFirstWeekdayOfMonth(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    return getFirstWeekdayOfMonth(dtDate);
} 

int getFirstWeekdayOfMonth(in d_time dtDate)
{
    d_time dtWkDate = d_time_nan;
    int    iMonth   = 0;
    int    iYear    = 0;
    
    iMonth = std.date.MonthFromTime(dtDate);
    iYear  = std.date.YearFromTime(dtDate);
    
    for ( int iDay = 1; iDay < 5; iDay++ )
    {
        dtWkDate = std.date.MakeDay(iYear, iMonth, iDay);
        
        if (isWeekday(dtWkDate) == true)
           return iDay;
    }    
} 

/+
 ' int getLastWeekdayOfMonth( in char[] = "" )
 '
 ' Returns the last weekday in the month of the given date.
 +/
int getLastWeekdayOfMonth(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
    
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    return getLastWeekdayOfMonth(dtDate);
} 

int getLastWeekdayOfMonth(in d_time dtDate)
{
    d_time dtWkDate = d_time_nan;
    int    iMonth   = 0;
    int    iYear    = 0;
    int    iLastDay = 0;
    
    iMonth   = std.date.MonthFromTime(dtDate);
    iYear    = std.date.YearFromTime(dtDate);
    iLastDay = getLastDayOfMonth(dtDate);
    
    for ( int iDay = iLastDay; iDay > iLastDay - 5; iDay-- )
    {
        dtWkDate = std.date.MakeDay(iYear, iMonth, iDay);
        
        if (isWeekday(dtWkDate) == true)
           return iDay;
    }       
} 

/+
 ' int getQuarterInYear( in char[] = "" )
 '
 ' Returns the quarter the given date falls into.
 +/
int getQuarterInYear(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;

    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());
    
    return getQuarterInYear(dtDate);
} 

int getQuarterInYear(in d_time dtDate)
{
    int iMonth   = 0;
    int iQuarter = 0;
    
    // std.date.MonthFromTime( dtDate ) returns as 0 thru 11
    iMonth = std.date.MonthFromTime(dtDate);
    
    switch (iMonth)
    {
      case  0,  1,  2: iQuarter = 1; break;
      case  3,  4,  5: iQuarter = 2; break;
      case  6,  7,  8: iQuarter = 3; break;
      case  9, 10, 11: iQuarter = 4; break;
    }
    
    return iQuarter;
} 

/+
 ' int inLeapYear( in char[] = "" )
 '
 ' Returns true if the given date is a leap year.
 +/
bool inLeapYear(in char[] sDate = ""c)
{
    d_time dtDate = d_time_nan;
 
    if (sDate != ""c)
        dtDate = std.date.parse(sDate);
    else
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    return (std.date.LeapYear(std.date.YearFromTime(dtDate)) == 1 ? true : false);
} 

/+
 ' int inLeapYear( in d_time = d_time_nan )
 '
 ' Returns true if the given date is a leap year.
 +/
bool inLeapYear(in d_time dtDate = d_time_nan)
{
    if (dtDate == d_time_nan)
        dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());

    return (std.date.LeapYear(std.date.YearFromTime(dtDate)) == 1 ? true : false);
} 

// Returns Local DateTime unless the string had a time that indicated otherwise.
d_time parse(in char[] s, out int iYear, out int iMonth, out int iDay)
{
    Date   dp;    
    d_time dtRDate = d_time_nan;
    d_time dtDay   = d_time_nan;
    d_time dtRTime = d_time_nan;

    dp.parse(s.dup);

    iYear  = dp.year;
    iMonth = dp.month;
    iDay   = dp.day;
     
    dtDay   = std.date.MakeDay(iYear, iMonth - 1, iDay);
    dtRTime = std.date.MakeTime(dp.hour, dp.minute, dp.second, dp.ms);
    dtRDate = dtDay * std.date.TicksPerDay + dtRTime;
                            
    return dtRDate;        
} 

unittest
{
    char[] sDate  = "07/04/1776"c;
    d_time dtDate = d_time_nan;
    int    iYear  = 0;
    int    iMonth = 0;
    int    iDay   = 0;
    
    writefln( "dateutils.calcAge( in char[] ).unittest"c );
    dtDate = std.date.UTCtoLocalTime(std.date.getUTCtime());
    iYear  = std.date.YearFromTime(dtDate);
    //assert( calcAge(sDate) == iYear - 1776 );

    writefln( "dateutils.isWeekday( in char[] ).unittest"c );
    assert( isWeekday(sDate) == true );

    writefln( "dateutils.isWeekend( in char[] ).unittest"c );
    assert( isWeekend(sDate) == false );

    writefln( "dateutils.getQuarterInYear( in char[] ).unittest"c );
    assert( getQuarterInYear(sDate) == 3 );

    writefln( "dateutils.getFirstWeekdayOfMonth( in char[] ).unittest"c );
    assert( getFirstWeekdayOfMonth(sDate) == 1 );

    writefln( "dateutils.getLastWeekdayOfMonth( in char[] ).unittest"c );
    assert( getLastWeekdayOfMonth(sDate) == 31 );

    writefln( "dateutils.getDayOfYear( in char[] ).unittest"c );
    assert( getDayOfYear(sDate) == 186 );

    writefln( "dateutils.inLeapYear( in char[] ).unittest"c );
    assert( inLeapYear(sDate) == true );

    writefln( "dateutils.getLastDayOfMonth( in char[] ).unittest"c );
    assert( getLastDayOfMonth(sDate) == 31 );

    writefln( "dateutils.getWeekOfYear( in char[] ).unittest"c );
    assert( getWeekOfYear(sDate) == 27 );

    writefln( "dateutils.parse( in char[], out int, out int, out int ).unittest"c );
    dtDate = .parse(sDate, iYear, iMonth, iDay);
    assert( iYear == 1776 );
    assert( iMonth == 7 );
    assert( iDay == 4 );
    assert( std.date.toUTCString(dtDate) == "Thu, 04 Jul 1776 00:00:00 UTC"c );
}

/+
 ' DATEDIFF
 ' -------------------------------------------------------------------------------------
 ' T-SQL / VB6 / Access (VBA) : DateDiff( Interval, StartDate, EndDate )
 '
 ' 1) Interval specifies on which part of the date to calculate the difference.
 ' 2) StartDate is the beginning date for the calculation.
 ' 3) EndDate is the ending date for the calculation.  
 '
 ' Returns the number of date and time boundaries crossed between two specified dates. 
 '
 ' Note: Compatiable with both MS SQL Server 2000's T-SQL version of DateDiff()
 '       and MS Visual Basic v6.0 / Access 2000 (VBA) version without the two 
 '       optional parameters. Date range allows is 01 Jan 100 to 31 Dec 9999.
 +/
long getDateDiff(in char[] sInterval, in d_time dtStartDate, in d_time dtEndDate)
{
    int isy = 0;
    int iey = 0;

    if (dtStartDate == d_time_nan || dtEndDate == d_time_nan)
        return d_time_nan;

    isy = std.date.YearFromTime(dtStartDate);
    iey = std.date.YearFromTime(dtEndDate);

    if ((isy < 100L) && (isy > 9999L)) 
        return d_time_nan;
    else if ((iey < 100L) && (iey > 9999L)) 
        return d_time_nan;

    switch (std.string.tolower(sInterval)) 
    {
      case "d"c, "dd"c, "day"c:
      {
          long lDays = roundToLong((dtEndDate - dtStartDate) / std.date.TicksPerDay);
          return (lDays > 1L ? lDays - 1L : lDays);
      }
      case "m"c, "mm"c, "month"c:
          return (iey - isy) * 12L +
                  std.date.MonthFromTime(dtEndDate) - std.date.MonthFromTime(dtStartDate);

      case "yy"c, "yyyy"c, "year"c:
          //return std.date.YearFromTime(dtEndDate) - std.date.YearFromTime(dtStartDate);
          return iey - isy;

      // Returns the number of weeks between the two dates...counts the number of Mondays,
      // starting with a Monday on the start date and ignore it if the end date is a Monday.
      case "w"c, "ww"c, "week"c:
      {
          d_time dtNewStartDate = dtStartDate;
          long   lWeeks         = 0L;
          long   lDayBypassed   = 0L;
        
          while (WeekDay(dtNewStartDate) != MONDAY)
          {
              dtNewStartDate += std.date.TicksPerDay;
              lDayBypassed++;
          }
        
          do 
          {
              if (WeekDay(dtNewStartDate) == MONDAY)
                  lWeeks++;
                
              dtNewStartDate += std.date.TicksPerDay;
            
          } while (dtNewStartDate < dtEndDate)
       
          lWeeks--;
          if (lDayBypassed > 1L)      
              lWeeks++;
           
          return lWeeks;
      }
      case "q"c, "qq"c, "quarter"c:
      {
          long lqq = ((iey  - isy) * 12L) +
                      (std.date.MonthFromTime(dtEndDate) - std.date.MonthFromTime(dtStartDate));
          
          return roundToLong(lqq / 3.0L);
      }
      case "hh"c, "hour"c:
          return roundToLong((dtEndDate - dtStartDate) / std.date.TicksPerHour);

      case "n"c, "mi"c, "minute"c:
          return roundToLong((dtEndDate - dtStartDate) / std.date.TicksPerMinute); 

      /+
      case "dy"c, "y"c, "dayofyear"c:
          break;
      +/

      case "s"c, "ss"c, "second"c:
          return roundToLong((dtEndDate - dtStartDate) / std.date.TicksPerSecond); 
 
      case "ms"c, "millsecond"c:
          return roundToLong(dtEndDate - dtStartDate);
      
      default:
          diehard("getDateDiff() not a supported interval!"c);
    }
} 

unittest
{
    writefln( "dateutils.getDateDiff( in char[], in d_time[], in d_time ).unittest"c );
    d_time dtEndDate = std.date.parse("05/23/1996"c);
    d_time dtStartDate;
    long lyy = 0L;
    long lmm = 0L;
    long ldd = 0L;
    long lww = 0L;
    long lqq = 0L;
    long lhh = 0L;
    long lme = 0L;
    long lss = 0L;
    long lms = 0L;

    dtStartDate = std.date.parse("10/01/1995"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 1 );
    assert( lmm == 7 );
    assert( ldd == 234 );
    assert( lww == 33 );
    assert( lqq == 2 );
    assert( lhh == 5640 );
    assert( lme == 338400 );
    assert( lss == 20304000 );
    //writefln("\"10/01/1995\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("09/01/1995"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 1 );
    assert( lmm == 8 );
    assert( ldd == 264 );
    assert( lww == 38 );
    assert( lqq == 3 );
    assert( lhh == 6360 );
    assert( lme == 381600 );
    assert( lss == 22896000 );
    //writefln("\"09/01/1995\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("08/01/1995"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 1 );
    assert( lmm == 9 );
    assert( ldd == 295 );
    assert( lww == 42 );
    assert( lqq == 3 );
    assert( lhh == 7104 );
    assert( lme == 426240 );
    assert( lss == 25574400 );
    //writefln("\"08/01/1995\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("07/01/1995"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 1 );
    assert( lmm == 10 );
    assert( ldd == 326 );
    assert( lww == 47 );
    assert( lqq == 3 );
    assert( lhh == 7848 );
    assert( lme == 470880 );
    assert( lss == 28252800 );
    //writefln("\"07/01/1995\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("01/01/1990"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 6 );
    assert( lmm == 76 );
    assert( ldd == 2333 );
    assert( lww == 333 );
    assert( lqq == 25 );
    assert( lhh == 56016 );
    assert( lme == 3360960 );
    assert( lss == 201657600 );
    //writefln("\"01/01/1990\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("01/02/1990"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 6 );
    assert( lmm == 76 );
    assert( ldd == 2332 );
    assert( lww == 333 );
    assert( lqq == 25 );
    assert( lhh == 55992 );
    assert( lme == 3359520 );
    assert( lss == 201571200 );
    //writefln("\"01/02/1990\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);

    dtStartDate = std.date.parse("12/31/1989"c);
    lyy = getDateDiff("year"c,    dtStartDate, dtEndDate);
    lmm = getDateDiff("month"c,   dtStartDate, dtEndDate);
    ldd = getDateDiff("day"c,     dtStartDate, dtEndDate);
    lww = getDateDiff("week"c,    dtStartDate, dtEndDate);
    lqq = getDateDiff("quarter"c, dtStartDate, dtEndDate);
    lhh = getDateDiff("hour"c,    dtStartDate, dtEndDate);
    lme = getDateDiff("minute"c,  dtStartDate, dtEndDate);
    lss = getDateDiff("second"c,  dtStartDate, dtEndDate);
    assert( lyy == 7 );
    assert( lmm == 77 );
    assert( ldd == 2334 );
    assert( lww == 333 );
    assert( lqq == 26 );
    assert( lhh == 56040 );
    assert( lme == 3362400 );
    assert( lss == 201744000 );
    //writefln("\"12/31/1989\", \"05/23/1996\", yy=%d, mm=%d, dd=%d, ww=%d, qq=%d, hh=%d, me=%d, ss=%d", 
    //         lyy, lmm, ldd, lww, lqq, lhh, lme, lss);
}

/+
 ' DATESERIAL
 ' --------------------------------------------------------
 ' VB6 / Access (VBA) : DateSerial( Year, Month, Day )
 '
 ' 1) Year is a number between 100 - 9999, inclusive, or a numeric expression. 
 ' 2) Month can be any number between -9999..-1 or 1..9999 inclusive, or a numeric expression.
 ' 3) Day can be any number between -9999..-1 or 1..9999, inclusive, or a numeric expression. 
 '
 ' Returns the date for a year, month and day passed in.
 '
 ' Note: Mostly compatiable with MS MS Visual Basic v6.0 / Access 2000 (VBA) version of DateSerial(), 
 '       the different being that zero isn't allowed (it means -1 in VBA), and that -1 isn't -2,
 '       and -2 isn't, etc as the VBA version does. Date range allows is 01 Jan 100 to 31 Dec 9999.
 +/
d_time getDateSerial(in long lYear, in long lMonth, in long lDay)
{
    d_time dtDate = d_time_nan;
    d_time dtDay  = d_time_nan;
    int    iyy    = 0;
    int    imm    = 0;
    int    idd    = 0;
    bool   bNeg   = false;
    
    // Zero is not allowed in any of the parameters.
    if (lYear == 0L || lMonth == 0L || lDay == 0L) 
        return d_time_nan;

    // Check ranges
    if ((lYear < 100L) && (lYear > 9999L)) 
        return d_time_nan;
    else if (lMonth < -9999L && lMonth > 9999L)
        return d_time_nan;
    else if (lDay < -9999L && lDay > 9999L)
        return d_time_nan;

    iyy = cast(int)lYear;
    imm = cast(int)lMonth;
    
    // if over 12 months, then add to the year
    if (imm >= 12 || imm <= -12)
    {
        iyy += roundToInteger(imm / 12);
        imm -= roundToInteger(imm / 12) * 12;
    }
    
    //writefln("iyy=%d, imm=%d", iyy, imm);
    
    // Adjust the neg month
    if (imm < 0)
    {
        iyy--;
        imm = (12 + imm);
    }
    //writefln("iyy=%d, imm=%d", iyy, imm);

    // Adjust days into months, and months into years
    idd = cast(int)lDay;
    if (idd < 0)
    {
        bNeg = true;
        imm--;

        while(idd < 0)
        {
            idd += getLastDayOfMonth(iyy, imm);
            //writefln("iyy=%d, imm=%d, idd=%d", iyy, imm, idd);
            imm--;
        }
        
        imm++;
    }
    else if (idd > 28)
    {
        imm++;
        
        while(idd > getLastDayOfMonth(iyy, imm))
        {
           idd -= getLastDayOfMonth(iyy, imm); 
           //writefln("iyy=%d, imm=%d, idd=%d", iyy, imm, idd);
           imm++;
        }
        
        imm--;
        
        if (imm >= 12 || imm <= -12)
        {
            iyy += roundToInteger(imm / 12);
            imm -= roundToInteger(imm / 12) * 12;
        }
    }

    // Adjust the days if on a leap-year
    if (std.date.LeapYear(iyy))
        if (bNeg)
            idd--;
        else    
            idd++;
       
    //writefln("iyy=%d, imm=%d, idd=%d", iyy, imm, idd); 
    
    // one last check on the year
    if ((lYear < 100L) && (lYear > 9999L)) 
        return d_time_nan;

    // Months in D are from 0 to 11, so subract 1   
    dtDay   = std.date.MakeDay(iyy, imm - 1, idd);
    dtDate  = std.date.MakeDate(dtDay, 0L);

    return dtDate;
} 

// In VB6's DateSerial() function, 0 equals -1, and -1 equals -2, and so on..., but this
//  version doesn't use 0, and so -1 equals -1, and -2 equals -2, and so on...
unittest
{
    writefln( "dateutils.getDateSerial( in long, in long, in long ).unittest"c );
    d_time dtRDate;

    // aws = 03/02/1997
    dtRDate = getDateSerial(2000, -29, -120);
    //writefln("getDateSerial(2000, -29, -120)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Sun, 02 Mar 1997 00:00:00 UTC"c );

    // getDateSerial(2000, 29, 1)   aws = 05/01/2002
    // getDateSerial(2000, 29, 120) aws = 08/28/2002
    dtRDate = getDateSerial(2000, 29, 120);
    //writefln("getDateSerial(2000, 29, 120)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Wed, 28 Aug 2002 00:00:00 UTC"c );    
    
    // getDateSerial(2000, 2, 29) aws = 2/29/2000 
    dtRDate = getDateSerial(2000, 2, 29);
    //writefln("getDateSerial(2000, 2, 29)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Wed, 01 Mar 2000 00:00:00 UTC"c );    

    // getDateSerial(2015, 2, 500) aws = 06/14/2016 
    dtRDate = getDateSerial(2015, 2, 500);
    //writefln("getDateSerial(2015, 2, 500)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Tue, 14 Jun 2016 00:00:00 UTC"c );
    
    // getDateSerial(2015, 2, 1) aws = 02/1/2015 
    dtRDate = getDateSerial(2015, 2, 1);
    //writefln("getDateSerial(2015, 2, 1)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Sun, 01 Feb 2015 00:00:00 UTC"c );    
    
    // getDateSerial(2015, 3, -1) aws = 02/29/2016 
    dtRDate = getDateSerial(2016, 3, -1);
    //writefln("getDateSerial(2016, 3, -1)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Mon, 29 Feb 2016 00:00:00 UTC"c );    

    // getDateSerial(2015, 3, -1) aws = 02/29/2016 
    dtRDate = getDateSerial(2016, -9999, -9999);
    //writefln("getDateSerial(2016, 3, -1)=%X, Date=\"%s\"", dtRDate, std.date.toUTCString(dtRDate));
    assert( std.date.toUTCString(dtRDate) == "Wed, 06 Apr 1155 00:00:00 UTC"c );    
}

/+
 ' DATEADD
 ' ---------------------------------------------------------------
 ' T-SQL / VB6 / Access (VBA) : DateAdd( Interval, Number, Date )
 '
 ' 1) Interval specifies on which part will be added to the date.
 '
 ' 2) Number is a number or a numeric expression that fall into 
 '    the ranges as noted below for each interval inclusive, 
 '    which are used to increment the interval. 
 '
 '    a) year: between 100 and 9999
 '    b) month: between -9999..-1 or 1..9999
 '    c) week: between -9999..-1 or 1..9999
 '    d) day: between -9999..-1 or 1..9999
 '    e) quarter: between -9999..-1 or 1..9999
 '    f) hour: between -9999..-1 or 1..9999
 '    g) minute: between -9999..-1 or 1..9999
 '    h) second: between -9999..-1 or 1..9999
 '    i) millisecond: between -9999..-1 or 1..9999
 '
 ' 3) Date is the base "Date" that the "Number" of "Interval" will be 
 '    added onto the returning the "Date" value.  
 '
 ' Returns a new datetime value based on adding an interval to the specified date. 
 '
 ' Note: Mostly compatiable with MS Visual Basic v6.0 / Access 2000 (VBA) 
 '       version of DateAdd(), the different being that zero isn't allowed
 '       (it means -1 in VBA), and that -1 isn't -2, and -2 isn't, etc as the
 '       VBA version does. Date range allows is 01 Jan 100 to 31 Dec 9999.
 +/
d_time getDateAdd(in char[] sInterval, in long lNumber, in d_time dtDate)
{
    int iyy = 0, imm = 0, idd = 0;
    int ihh = 0, ime = 0, iss = 0, ims = 0;
    int idy = 0;

    if (lNumber == 0L)
        return dtDate;
    else if (dtDate == d_time_nan)
        return d_time_nan;
    else if (sInterval == "yy"c || sInterval == "yyyy"c || sInterval == "year"c)
        if ((lNumber < 100L) && (lNumber > 9999L)) 
            return d_time_nan;
    else if (lNumber < -9999L && lNumber > 9999L)
        return d_time_nan;

    // Gather all the date and time info 
    // for rebuilding the return date.
    iyy = std.date.YearFromTime(dtDate);
    imm = std.date.MonthFromTime(dtDate);
    idd = std.date.DateFromTime(dtDate); 
    ihh = std.date.HourFromTime(dtDate);
    ime = std.date.MinFromTime(dtDate);
    iss = std.date.SecFromTime(dtDate);
    ims = std.date.msFromTime(dtDate);
    
    switch (std.string.tolower(sInterval)) 
    {
      case "d"c, "dd"c, "day"c:
      {
          idd += lNumber; 
          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims); 
      }
      case "m"c, "mm"c, "month"c:
      {
          imm += lNumber;
          idy  = 0;

          if (imm >= 12)
          {
              iyy += roundToInteger(imm / 12);
              imm -= roundToInteger((imm / 12) * 12);
          }

          idy = getLastDayOfMonth(iyy, imm);
          if (idd > idy)
              idd = idy;

          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims);  
      }
      case "yy"c, "yyyy"c, "year"c:
      {
          iyy += lNumber;
          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims); 
      }
      case "w"c, "ww"c, "week"c:
      {   
          idd += lNumber * 7;
          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims);  
      }
      case "q"c, "qq"c, "quarter"c:
      {
          imm += lNumber * 3;
          idy  = 0;

          if (imm >= 12)
          {
              iyy += roundToInteger(imm / 12);
              imm -= roundToInteger((imm / 12) * 12);
          }

          idy = getLastDayOfMonth(iyy, imm);
          if (idd > idy)
              idd = idy;

          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims); 
      }
      case "hh"c, "hour"c:
      {
          ihh += lNumber;

          if (ihh >= 24)
          {
              idd += roundToInteger(ihh / 24);
              ihh -= roundToInteger((ihh / 24) * 24);
          }

          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims); 
      }
      case "n"c, "mi"c, "minute"c:
      {
          ime += lNumber;

          if (ime >= 60)
          {
              ihh += roundToInteger(ime / 60);
              ime -= roundToInteger((ime / 60) * 60);
          }

          if (ihh >= 24)
          {
              idd += roundToInteger(ihh / 24);
              ihh -= roundToInteger((ihh / 24) * 24);
          }

          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims);   
      }
      case "s"c, "ss"c, "second"c:
      {
          iss += lNumber;
          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims);       
      }
      case "ms"c, "millsecond"c:
      {
          ims += lNumber;
          return createNewDateTime(iyy, imm, idd, ihh, ime, iss, ims);     
      }
      default:
          diehard("getDateAdd() not a supported interval!"c);
    }
} 

unittest
{
    writefln( "dateutils.getDateAdd( in char[], in long, in d_time ).unittest"c );
    d_time dtDate  = d_time_nan;
    d_time dtRDate = d_time_nan;

    dtDate = std.date.parse("12/31/2008 13:25:05 UTC"c); 

    // Add two years to the current date
    dtRDate = getDateAdd("year"c, 2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Fri, 31 Dec 2010 13:25:05 UTC"c);

    // Subract two years to the current date
    dtRDate = getDateAdd("year"c, -2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Sun, 31 Dec 2006 13:25:05 UTC"c );
 
    // Add two months to the current date
    dtRDate = getDateAdd("month"c, 2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Sat, 28 Feb 2009 13:25:05 UTC"c );

    // Subract months to the current date
    dtRDate = getDateAdd("month"c, -2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Fri, 31 Oct 2008 13:25:05 UTC"c );

    // Add two days to the current date
    dtRDate = getDateAdd("day"c, 2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Fri, 02 Jan 2009 13:25:05 UTC"c );

    // Subract two days to the current date
    dtRDate = getDateAdd("day"c, -2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Mon, 29 Dec 2008 13:25:05 UTC"c );

    // Add two days to the current date
    dtDate = std.date.parse("02/29/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("day"c, 2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Thu, 02 Mar 2000 22:47:55 UTC"c );

    // Add one month to the current date (to Feb 29 in a leap year)
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("month"c, 1, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Tue, 29 Feb 2000 22:47:55 UTC"c );

    // Subract one month to the current date (to Feb 29 in a leap year)
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("month"c, -1, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Wed, 29 Dec 1999 22:47:55 UTC"c );

    // Add five hours to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("hour"c, 5, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Tue, 01 Feb 2000 03:47:55 UTC"c );

    // Subract five hours to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("hour"c, -5, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Mon, 31 Jan 2000 17:47:55 UTC"c );
 
    // Add 122 minutes (two hours and 2 minutes worth) to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("minute"c, 122, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Tue, 01 Feb 2000 00:49:55 UTC"c );

    // Subract 122 minutes (two hours and 2 minutes worth) to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("minute"c, -122, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Mon, 31 Jan 2000 20:45:55 UTC"c );

    // Add one day and one second worth of seconds to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("second"c, (24 * 60 * 60) + 1, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Tue, 01 Feb 2000 22:47:56 UTC"c );

    // Subract one day and one second worth of seconds to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("second"c, (-24 * 60 * 60) - 1, dtDate);
    assert(std.date.toUTCString(dtRDate) == "Sun, 30 Jan 2000 22:47:54 UTC"c );

    // Add one day and one second worth of milliseconds to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("millsecond"c, (24 * 60 * 60 * 1000) + 1000, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Tue, 01 Feb 2000 22:47:56 UTC"c );

    // Subract one day and one second worth of milliseconds to the current date
    dtDate = std.date.parse("01/31/2000 22:47:55 UTC"c);
    dtRDate = getDateAdd("millsecond"c, (-24 * 60 * 60 * 1000) - 1000, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Sun, 30 Jan 2000 22:47:54 UTC"c );

    // Add two quarters (each quarter equals three months) worth of months to the current date
    dtDate = std.date.parse("12/14/2007 18:12:25 UTC"c); 
    dtRDate = getDateAdd("quarter"c, 2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Sat, 14 Jun 2008 18:12:25 UTC"c );

    // Subract two quarters (each quarter equals three months) worth of months to the current date
    dtDate = std.date.parse("12/14/2007 18:12:25 UTC"c); 
    dtRDate = getDateAdd("quarter"c, -2, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Thu, 14 Jun 2007 18:12:25 UTC"c );

    // Add five weeks (each week is equals to seven days) worth of days to the current date
    dtDate = std.date.parse("02/14/2008 21:35:57 UTC"c); 
    dtRDate = getDateAdd("week"c, 5, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Thu, 20 Mar 2008 21:35:57 UTC"c );

    // Subract five weeks (each week is equals to seven days) worth of days to the current date
    dtDate = std.date.parse("02/14/2008 21:35:57 UTC"c); 
    dtRDate = getDateAdd("week"c, -5, dtDate);
    assert( std.date.toUTCString(dtRDate) == "Thu, 10 Jan 2008 21:35:57 UTC"c );

    //writefln("minute: dtDate=\"%s\", dtRDate=\"%s\"", 
    //          std.date.toUTCString(dtDate), std.date.toUTCString(dtRDate));
}

/+
 ' DATEPART
 ' --------------------------------------------------------
 ' T-SQL / VB6 / Access (VBA) : DatePart( Interval, Date )
 '
 ' 1) Interval specifies on which part of the date to return.
 ' 2) Date is the date to pull data from.
 '
 ' Returns the interval specifies for the given date passed in. 
 '
 ' Note: Compatiable with both MS SQL Server 2000's T-SQL version of 
 '       DatePart() and MS Visual Basic v6.0 / Access 2000 (VBA) version.
 +/
long getDatePart(in char[] sInterval, in d_time dtDate)
{
    if (dtDate == d_time_nan)
        return d_time_nan;

    switch (std.string.tolower(sInterval)) 
    {
      case "d"c, "dd"c, "day"c:
          return std.date.DateFromTime(dtDate);
          
      case "m"c, "mm"c, "month"c:
          return std.date.MonthFromTime(dtDate) + 1;

      case "yy"c, "yyyy"c, "year"c:
          return std.date.YearFromTime(dtDate);

      case "w"c, "ww"c, "week"c:
          return getWeekOfYear(dtDate);
          
      case "q"c, "qq"c, "quarter"c:
          return getQuarterInYear(dtDate);
          
      case "hh"c, "hour"c:
          return std.date.HourFromTime(dtDate);

      case "n"c, "mi"c, "minute"c:
          return std.date.MinFromTime(dtDate); 

      case "dy"c, "y"c, "dayofyear"c:
          return getDayOfYear(dtDate);

      case "dw"c, "weekday"c:
          return std.date.WeekDay(dtDate);
         
      case "s"c, "ss"c, "second"c:
          return std.date.SecFromTime(dtDate); 
 
      case "ms"c, "millsecond"c:
          return std.date.msFromTime(dtDate);
      
      default:
          diehard("getDateDiff() not a supported interval!"c);
    }
} // end getDatePart()

unittest
{
    writefln( "dateutils.getDatePart( in char[], in d_time ).unittest"c );
    d_time dtDate = d_time_nan;

    dtDate = std.date.parse("12/31/2008 13:25:05 UTC"c ); 
    assert( getDatePart("year"c, dtDate) == 2008 );
    assert( getDatePart("month"c, dtDate) == 12 );
    assert( getDatePart("day"c, dtDate) == 31 );
    assert( getDatePart("hour"c, dtDate) == 13 );
    assert( getDatePart("minute"c, dtDate) == 25 );
    assert( getDatePart("second"c, dtDate) == 5 );
    assert( getDatePart("millsecond"c, dtDate) == 0 );
    assert( getDatePart("week"c, dtDate) == 53 );
    assert( getDatePart("quarter"c, dtDate) == 4 );
    assert( getDatePart("dayofyear"c, dtDate) == 366 );
    assert( getDatePart("weekday"c, dtDate) == WEDNESDAY );
    
    dtDate = std.date.parse("07/15/2015 23:11:55 UTC"c); 
    assert( getDatePart("year"c, dtDate) == 2015 );
    assert( getDatePart("month"c, dtDate) == 7 );
    assert( getDatePart("day"c, dtDate) == 15 );
    assert( getDatePart("hour"c, dtDate) == 23 );
    assert( getDatePart("minute"c, dtDate) == 11 );
    assert( getDatePart("second"c, dtDate) == 55 );
    assert( getDatePart("millsecond"c, dtDate) == 0 );
    assert( getDatePart("week"c, dtDate) == 29 );
    assert( getDatePart("quarter"c, dtDate) == 3 );
    //writefln("dy=%d", getDatePart("dayofyear", dtDate));
    //writefln("getDatePart(\"dayofyear\", dtDate)=\"%s\"", toUTCString(dtDate));
    assert( getDatePart("dayofyear"c, dtDate) == 196 );
    assert( getDatePart("weekday"c, dtDate) == WEDNESDAY );    
}

d_time getDateOnly(in d_time dtDate)
{
    d_time dtRDate = d_time_nan;
    int iyy = 0, imm = 0, idd = 0;
   
    iyy = std.date.YearFromTime(dtDate);
    imm = std.date.MonthFromTime(dtDate);
    idd = std.date.DateFromTime(dtDate); 

    dtRDate = std.date.MakeDay(iyy, imm, idd);
    dtRDate = std.date.MakeDate(dtRDate, 0L); 
    
    return dtRDate;
} 

// Note that zero date becomes 01/01/1970, but it can be ignored,
// use the toUTCTimeString() function to see only the time as a string.
d_time getTimeOnly(in d_time dtDate)
{
    d_time dtRTime = d_time_nan;
    int ihh = 0, ime = 0, iss = 0, ims = 0;
   
    ihh = std.date.HourFromTime(dtDate);
    ime = std.date.MinFromTime(dtDate);
    iss = std.date.SecFromTime(dtDate);
    ims = std.date.msFromTime(dtDate);

    dtRTime = std.date.MakeTime(ihh, ime, iss, ims);
    
    return dtRTime;
}

// Please note, that month is passed in as 0 thru 11, instead of 1 thru 12.
// Also use the toUTCString() function to see the original date as a string.
d_time createNewDateTime
(
    in int iyy,
    in int imm,
    in int idd,
    in int ihh,
    in int ime,
    in int iss,
    in int ims,
    in int imin = 100,
    in int imax = 9999
)
{
    d_time dtRDate = d_time_nan;
    d_time dtRTime = d_time_nan;
    
    if (iyy < imin || iyy > imax)
        return d_time_nan;

    dtRDate = std.date.MakeDay(iyy, imm, idd);
    dtRTime = std.date.MakeTime(ihh, ime, iss, ims);
    dtRDate = std.date.MakeDate(dtRDate, dtRTime);

    if (iyy >= imin || iyy <= imax)
        return dtRDate;
    else
        return d_time_nan;
}
    
// Modifed version of Walter's std.date toUTCString() function
// for showing only the time piece of the d_time variable.
char[] toUTCTimeString(d_time t)
{
    // Years are supposed to be -285616 .. 285616, or 7 digits
    // "Tue, 02 Apr 1996 02:04:57 GMT"
    char[] buffer = new char[8 + 7 + 1];
    int len;

    if (t == d_time_nan)
        return "Invalid Date"c;

    len = std.c.stdio.sprintf(cast(char *)buffer, 
                  cast(char *)"%02d:%02d:%02d UTC"c,
                  cast(int)std.date.HourFromTime(t), 
                  cast(int)std.date.MinFromTime(t), 
                  cast(int)std.date.SecFromTime(t));

    // Ensure no buggy buffer overflows
    assert( len < buffer.length );

    return buffer[0..len];
}

private void diehard(in char[] s)
{
    throw new Exception(s);
}

debug( dateutils )
{
int main()
{
    writefln( "dateutils.unittest finished!"c );
/+    
    d_time dtDate1, dtDate2;
    dtDate1 = std.date.UTCtoLocalTime(parse("01/31/2010 23:50:55"));
    //dtDate2 = std.date.UTCtoLocalTime(std.date.getUTCtime());
    
    writefln("parse only=\"%s\" \"%s\"", 
             toDateString(parse("01/31/2010 23:50:55 UTC")), 
             toTimeString(parse("01/31/2010 23:50:55 UTC")));
             
    writefln("UTCtoLocalTime=\"%s\" \"%s\"", toDateString(dtDate1), toTimeString(dtDate1));
    //writefln("UTCtoLocalTime=\"%s\"", dtDate1);
    +/
    return 0;
    
} // end int main()    
}
