/+
 ' File / Module   : stringplus.d  
 ' Version         : v0.14 beta
 ' Author(s)       : David L. 'SpottedTiger' Davis
 '                 : Walter Bright's std.string.d used as the foundation
 '                 : for some of these functions.
 ' Source Platform : Intel x86 / WinXP SP2 
 ' Target Platform : Should work on any CPU / OS combo that D compiles on.
 '                 :
 ' Created Date    : 16.Oct.04 Gathered up the extended functions 
 '                 : written as string.d extra functions.
 '                 : Complied and Tested with dmd v0.102
 '                 :
 ' Modified Date   : 24.Nov.04 Added the isDigits() and addOrdSuffix() functions,
 '                 :           recomplied and tested with dmd v0.106.
 '                 : 30.Nov.04 Recomplied and Tested with dmd v0.108.
 '                 : 22.Jan.05 Recomplied and Tested with dmd v0.111.
 '                 : 20.Mar.05 Recomplied and Tested with dmd v0.119.
 '                 : 17.May.05 Added both xtoul() and ultox() functions, plus 
 '                 :           made a correction to the addOrdSuffix() function,
 '                 :           and recomplied and retested with dmd v0.123.
 '                 : 18.May.05 Corrected ultox() to handle a zero, duh! 
 '                 :           How did I miss that one? Added the isHex() 
 '                 :           function.
 '                 : 10.Jun.05 Recomplied and Tested with dmd v0.126.
 '                 : 17.Sep.05 Recomplied and Tested with dmd v0.131.
 '                 : 24.Oct.05 Compiled and tested with dmd v0.137.
 '                 : 30.Dec.05 Compiled and tested with dmd v0.142.
 '                 : 22.Mar.06 Compiled and tested with dmd v0.150.
 '                 : 28.May.06 Recompiled and Tested with dmd v0.159.
 '                 : 19.Jul.06 All "Imports" are now private by default  
 '                 :           (which shouldn't be a problem, since I  
 '                 :           always use the private keyword) in 
 '                 :           dmd v0.163, recompiled and tested.
 '                 : 23.Oct.06 Corrected one "shadowing declaration" error,
 '                 :           then compiled and tested with dmd v0.172.
 '                 : 05.Nov.06 Recompiled and Tested with dmd v0.173.
 '                 : 09.Dec.06 Recompiled and Tested with dmd v0.177.
 '                 : 05.Jan.07 Recompiled and Tested with dmd v1.0.
 '                 :
 ' Requirements    : Must complie with dmd v0.92+ and above, because
 '                 : default parameters are being used. 
 '                 :
 ' Licence         : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: Over time I hope to add more functions, and to better  
 '       fine tune the existing ones more.
 '
 ' To complie for a unittest:
 ' dmd stringplus.d -debug=stringplus -unittest
 '
 ' To compile a D program that use stringplus.d:
 ' dmd MySource.d stringplus.d
 '
 ' 01) char[] ireplace( in char[], in char[], in char[] )
 ' 02) int icount( in char[], in char[] )
 ' 03) int ifindx( in char[], in char[], in int = 0 )
 ' 04) int irfindx( in char[], in char[], in int = 0 )
 ' 05) int findx( in char[], in char[], in int = 0 )
 ' 06) int rfindx( in char[], in char[], in int = 0 )
 ' 07) bool isRomanNumeral( in char )
 ' 08) bool isRomanNumeral( in char[] )
 ' 09) bool isASCIIString( in char[], bool = false )
 ' 10) char[] asciiProperCase( in char[] )
 ' 11) char[] toSoundex( in char[] )
 ' 12) bool isDigits( in char[] )
 ' 13) char[] addOrdSuffix( in ulong )
 ' 14) ulong xtoul( in char[] )
 ' 15) char[] ultox( in ulong )
 ' 16) bool isHex( in char[] )
 ' 
 +/
module stringplus;

private import std.stdio;   // Needed for writefln();
private import std.string;  
private import std.ctype;   // for tolower( dchar ) and toupper( dchar );

const char[]   WHITESPACE = ".,-\x20\'\t\v\f\n\r"c;
const char[14] ROMAN      = "IVXLCDMivxlcdm"c;

/+
 ' Gathered the 1000 most commonly used English words 
 ' together from the web, and then got a character count 
 ' on each. Which was then used to reordered the characters 
 ' from the most used, to the least used.
 '
 ' e=638, t=385, a=370, r=369, o=354, n=300,
 ' s=290, i=280, l=270, c=197, h=182, d=175,
 ' u=149, p=143, m=126, g=118, w=099, f=088,
 ' b=087, y=081, v=052, k=051, x=014, q=011,
 ' j=008, z=001
 '
 +/
const char[26] LOWERCASE = "etaronsilchdupmgwfbyvkxqjz"c;
const char[26] UPPERCASE = "ETARONSILCHDUPMGWFBYVKXQJZ"c; 

/+
 ' Function        : char[] asciiProperCase( in char[] )
 ' Author          : David L. 'SpottedTiger' Davis
 ' Compiler        : dmd v0.92
 ' Date Created    : 03.Jun.04 
 ' Date Modified   : 24.Aug.04 Compiled and tested with dmd v0.99
 ' Licence         : Pubic Domain
 ' Requirements    : std.string
 ' ------------------------------------------------------------------
 ' Purpose : This function was created to handle Names and Addresses.
 '
 ' 01) Caps the first character of each word.
 ' 02) Converts o'bein to O'Bein
 ' 03) Converts apple-bees to Apple-Bees
 ' 04) Ignores commas, ex: doe,jane to Doe,Jane
 ' 05) Coverts mcdonald to McDonald
 ' 06) Converts all roman numerials to caps
 ' 07) Converts 1ST, 2ND, 3RD, 4TH to 1st, 2nd, 3rd, 4th
 ' 08) Trims off leading and trailing spaces
 ' 09) Converts po and pob to PO and POB
 ' 10) Converts p.o. to P.O. or caps the next character that follow a period
 '
 +/
char[] asciiProperCase(in char[] sStrPassed)
{
    const char[] WHITESPACE = ".,-\x20\'\t\v\f\n\r"c;
        
    bool   bWhitespaceFound = false;
    bool   bRomanFound      = false;
    int    iStrPosWK        = 0;
    int    iStrPos          = 0;
    char[] sStr             = sStrPassed.dup; //new char[ sStrPassed.length ]; 
    
    //sStr = sStrPassed; 
    
    if (sStr.length > 0) 
    {
        /*
          Initial pass:
          ---------------
          1) strip all leading and trailing spaces
          2) lowercase all characters
          3) very 1st character set to uppercase
          4) capitalize each 1st character in every word
        */

        sStr = std.string.capitalize(std.string.tolower(std.string.strip(sStr)));

        //sStr = capwords( sStr ); // not used because it does too much for my needs

        foreach(int iStrPos2, char cChar; sStr)
        { 
            if (std.string.find(WHITESPACE, cChar) != -1)     
            { 
                bWhitespaceFound = true;
            } 
            else if (bWhitespaceFound == true) 
            { 
                bWhitespaceFound = false;
                sStr[iStrPos2] = std.ctype.toupper(sStr[iStrPos2]); 
            } 

            /+*********************
             *  Roman Numeral(s)  *
             *********************+/   
            bRomanFound = isRomanNumeral(sStr[iStrPos2]);
            iStrPosWK = iStrPos2;
    
            while (iStrPosWK < sStr.length && bRomanFound == true)
            {
                //writefln("Roman Number found, sStr.length=%d, iStrPosWK=%d", sStr.length, iStrPosWK );
                iStrPosWK++;
        
                if (iStrPosWK < sStr.length) 
                    bRomanFound = isRomanNumeral(sStr[iStrPosWK]);
            } 
    
            if (iStrPosWK >= sStr.length) 
                iStrPosWK = sStr.length - 1;

            if ((std.string.find(WHITESPACE, sStr[iStrPosWK]) != -1 || 
                iStrPosWK == sStr.length - 1)) 
            {
                // Handles before WHITESPACE
                bRomanFound = isRomanNumeral(sStr[(iStrPosWK - 2 >= 0) ? iStrPosWK - 2 : iStrPosWK]);
        
                if (bRomanFound == true)
                {      
                    sStr[iStrPos2 .. iStrPosWK] = std.string.toupper(sStr[iStrPos2 .. iStrPosWK]);
            
                    // Needed for the very last character if it's a Roman Numeral
                    if (iStrPosWK == sStr.length - 1) 
                        sStr[iStrPosWK] = std.ctype.toupper(sStr[iStrPosWK]);
            
                    iStrPos2 = iStrPosWK;
                    bWhitespaceFound = true;
                }     
            } 
    
            //writefln("Char=%s", sStr[ iStrPos2 ]);

        } // end-foreach

        /+
         ' * Work on all the special cases *
         '
         ' Need a find with replace for the following:
         ' 1) po     -> PO
         ' 2) pob    -> POB
         ' 3) mccain -> McCain
         '
         +/

        // Handles " po "
        iStrPos = 0;

        if (std.string.icmp(sStr[0 .. 3], "PO\x20"c) == 0) 
            sStr[0 .. 3] = "PO\x20"c;

        while (iStrPos < sStr.length - 1 && iStrPos != -1)
        {
            iStrPos = ifindx(sStr, "\x20PO\x20"c, iStrPos);

            if (iStrPos != -1) 
            {
                sStr[iStrPos .. iStrPos + 4] = "\x20PO\x20"c;
                iStrPos = iStrPos + 5;
            } 
        } 

        // Handles " pob " 
        iStrPos = 0;

        if (std.string.icmp(sStr[0 .. 4], "POB\x20"c) == 0) 
            sStr[0 .. 4] = "POB\x20"c;

        while (iStrPos < sStr.length - 1 && iStrPos != -1)
        {
            iStrPos = ifindx(sStr, "\x20POB\x20"c, iStrPos);

            if (iStrPos != -1) 
            {
                sStr[iStrPos .. iStrPos + 5] = "\x20POB\x20"c;
                iStrPos = iStrPos + 6;
            } 
        } 

        // Handles " Mc"
        iStrPos = 0;

        if (std.string.icmp(sStr[0 .. 2], "MC"c) == 0 && 
            std.string.find(WHITESPACE, sStr[2 .. 3]) == -1) 
        {
            sStr[0 .. 2] = "Mc"c;
            sStr[2] = std.ctype.toupper(sStr[2]);
        } 

        while (iStrPos < sStr.length - 1 && iStrPos != -1)
        {
            iStrPos = ifindx(sStr, "\x20MC"c, iStrPos);

            if (iStrPos != -1) 
            {
                if (iStrPos + 4 < sStr.length)
                { 
                    sStr[iStrPos .. iStrPos + 3] = "\x20Mc"c;
                    sStr[iStrPos + 3] = std.ctype.toupper(sStr[iStrPos + 3]);
                    iStrPos = iStrPos + 5;
                } 
            } 
        } 
    } 

    return sStr;
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.asciiProperCase( in char[] ).unittest"c );
    
    char[] s1 = "mccain c apPlebEe 123RD p.o. po pob bOx mcgally 32 aNy-PLAce ivx souTH sT aNy'citY,pa usa mmcxv "c;
    char[] s2;

    s2 = asciiProperCase(s1);
    assert( s2 != s1 );
    assert( s2 == "McCain C Applebee 123rd P.O. PO POB Box McGally 32 Any-Place IVX South St Any'City,Pa Usa MMCXV"c );
    
    //writefln( "APPLE=\"%s\", ORANGE=\"%s\"", asciiProperCase( "APPLE" ), asciiProperCase( "ORANGE" ) );
}   

/+
 ' Function      : int ifindx( in char[], in char[], in int = 0 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Complied      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set 
 '                           the third parameter as a default of 0
 '               : 24.Aug.04 Now that ifind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from ifind() to ifindx() 
 '                           [ meaning ifind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 ' Requirements  : std.string
 ' Licence       : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: Meant to extend the Non-Case Sensitive version of ifind() with 
 '       an optional start searching from here string position parameter.
 '
 +/
int ifindx(in char[] sStr, in char[] sSubStr, in int iStartPos = 0)
{
    char[] sTmpStr;
    int    iRtnVal;
    int    iStrLen = sStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || sSubStr.length < 1) 
        return -1;
    
    // If greater than to upper boundary return not found    
    if (iStartPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iStartPos < 0) 
        return - 1; 
           
    sTmpStr = sStr[iStartPos .. iStrLen];
  
    if (iStartPos == 0) 
        return std.string.ifind(sTmpStr, sSubStr);   
    else
    {
        iRtnVal = std.string.ifind(sTmpStr, sSubStr);

        return (iRtnVal != -1) ? iStartPos + iRtnVal : iRtnVal;
    } 
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.ifindx( in char[], in char[], in int = 0 ).unittest"c );
    
    char[] sPlts = "Mars: the fourth Rock (Planet) from the Sun."c;
    char[] sMars = "Who\'s \'My Favorite Maritian?\'"c;
    int    i;

    i = ifindx(null, "a"c);
    assert( i == -1 );
    i = ifindx("def"c, "a"c);
    assert( i == -1 );
    i = ifindx("abbA"c, "a"c, 1);
    assert( i == 3 );
    i = ifindx("deF"c, "f"c);
    assert( i == 2 );
    i = ifindx("dfefffg"c, "fFf"c);
    assert( i == 3 );
    i = ifindx("dfeffgfff"c, "FFF"c, 4);
    assert( i == 6 );
    
    i = ifindx(sMars, "MY fAVe"c);
    assert( i == -1 );
    i = ifindx(sMars, "mY fAVOriTe"c, 3);
    assert( i == 7 );
    i = ifindx(sPlts, "mArS:"c);
    assert( i == 0 );
    i = ifindx(sPlts, "rOcK"c);
    assert( i == 17 );
    i = ifindx(sPlts, "rOcK"c, 19);
    assert( i == -1 );
    i = ifindx(sPlts, "Un."c, 35);
    assert( i == 41 );
    i = ifindx(sPlts, sPlts);
    assert( i == 0 );
}

/+
 ' Function      : int irfindx( in char[], in char[], in int = -1 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Compiled      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set
 '                           the third parameter as a default of -1
 '               : 24.Aug.04 Now that irfind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from irfind() to irfindx() 
 '                           [ meaning irfind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 ' Requirements  : std.string
 ' Licence       : Public Domain
 ' --------------------------------------------------------------------------
 ' Note: Meant to extend the Non-Case Sensitive version of irfind() with 
 '       an optional start searching from here string position parameter.
 '
 +/
int irfindx(in char[] sStr, in char[] sSubStr, in int iEndPos = -1)
{
    char[] sTmpStr;
    int    iStrLen = sStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || sSubStr.length < 1) 
        return -1;

    // If iEndPos == -1 get the full length of the string
    if (iEndPos == -1) 
        iEndPos = iStrLen - 1;
        
    // If greater than to upper boundary return not found    
    else if (iEndPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iEndPos < 0) 
        return -1;      
    
    sTmpStr = sStr[0 .. iEndPos + 1];

    if (sTmpStr.length < sSubStr.length)
        return -1;
        
    return std.string.irfind(sTmpStr, sSubStr);
} 

unittest
{
    debug( stringplus ) writefln("stringplus.irfindx( in char[], in char[], in int = -1 ).unittest"c );

    char[] sPlts = "Mars: the fourth Rock (Planet) from the Sun."c;
    char[] sMars = "Who\'s \'My Favorite Maritian?\'"c;
    int     i;

    i = irfindx("abcdefcdef"c, "DEF"c, 8);
    assert( i == 3 );
    i = irfindx("abcdefcdef"c, "Cd"c, 6);
    assert( i == 2 );
    
    i = irfindx("abcdefcdef"c, "c"c, 5);
    assert( i == 2 );
    i = irfindx("abcdefcdef"c, "x"c);
    assert( i == -1 );
    i = irfindx("abcdefcdef"c, "xY"c);
    assert( i == -1 );
    i = irfindx("abcdefcdef"c, ""c);
    assert( i == -1 );
    
    i = irfindx(sMars, "RiTE maR"c);
    assert( i == 14 );
    i = irfindx(sPlts, "FOuRTh"c);
    assert( i == 10 );
    i = irfindx(sMars, "whO\'s \'MY"c);
    assert( i == 0 );
    i = irfindx(sMars, sMars);
    assert( i == 0 );
}    

/+
 ' Function      : int findx( in char[], in char[], in int = 0 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Complied      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set 
 '                           the third parameter as a default of 0
 '               : 24.Aug.04 Now that ifind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from ifind() to ifindx() 
 '                           [ meaning ifind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 ' Requirements  : std.string
 ' Licence       : Public Domain
 ' ---------------------------------------------------------------------------
 ' Note: Meant to extend the Case Sensitive version of ifind() with 
 '       an optional start searching from here string position parameter.
 '
 +/
int findx(in char[] sStr, in char[] sSubStr, in int iStartPos = 0)
{
    char[] sTmpStr;
    int    iRtnVal;
    int    iStrLen = sStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || sSubStr.length < 1) 
        return -1;
    
    // If greater than to upper boundary return not found    
    if (iStartPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iStartPos < 0) 
        return - 1; 
           
    sTmpStr = sStr[iStartPos .. iStrLen];
  
    if (iStartPos == 0) 
        return std.string.find(sTmpStr, sSubStr);   
    else
    {
        iRtnVal = std.string.find(sTmpStr, sSubStr);

        return (iRtnVal != -1) ? iStartPos + iRtnVal : iRtnVal;
    } 
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.findx( in char[], in char[], in int = 0 ).unittest"c );
    
    char[] sPlts = "Mars: the fourth Rock (Planet) from the Sun."c;
    char[] sMars = "Who\'s \'My Favorite Maritian?\'"c;
    int    i;

    i = findx(null, "a"c);
    assert( i == -1 );
    i = findx("def"c, "a"c);
    assert( i == -1 );
    i = findx("abbA"c, "a"c, 0);
    assert( i == 0 );
    i = findx("deF"c, "f"c);
    assert( i == -1 );
    i = findx("dfefFfg"c, "fFf"c);
    assert( i == 3 );
    i = findx("dfeffgfff"c, "FFF"c, 4);
    assert( i == -1 );
    
    i = findx(sMars, "MY fAVe"c);
    assert( i == -1 );
    i = findx(sMars, "mY fAVOriTe"c, 3);
    assert( i == -1 );
    i = findx(sPlts, "Mars:"c);
    assert( i == 0 );
    i = findx(sPlts, "Rock"c);
    assert( i == 17 );
    i = findx(sPlts, "Rock"c, 19);
    assert( i == -1 );
    i = findx(sPlts, "un."c, 35);
    assert( i == 41 );
    i = findx(sPlts, sPlts);
    assert( i == 0 );
}

/+
 ' Function      : int rfindx( in char[], in char[], in int = -1 )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Compiled      : DigitalMars "D" aka Mars v0.92
 ' Created Date  : 03.Jun.04 
 ' Modified Date : 08.Jun.04 Removed the wrapper function and set
 '                           the third parameter as a default of -1
 '               : 24.Aug.04 Now that irfind() is a part of Phobos,
 '                           and the fact that I wrote this original
 '                           function to use a default...I've renamed
 '                           this function from irfind() to irfindx() 
 '                           [ meaning irfind() eXtended :)) ]
 '               : 24.Aug.04 Compiled and tested with dmd v0.99
 '               : 01.Sep.04 Compiled and tested with dmd v0.101
 ' Requirements  : std.string
 ' Licence       : Public Domain
 ' --------------------------------------------------------------------------
 ' Note: Meant to extend the Case Sensitive version of rfind() with 
 '       an optional start searching from here string position parameter.
 '
 +/
int rfindx(in char[] sStr, in char[] sSubStr, in int iEndPos = -1)
{
    char[] sTmpStr;
    int    iStrLen = sStr.length;
    
    // If either of the string parameters are empty, return not found
    if (iStrLen < 1 || sSubStr.length < 1) 
        return -1;

    // If iEndPos == -1 get the full length of the string
    if (iEndPos == -1) 
        iEndPos = iStrLen - 1;
        
    // If greater than to upper boundary return not found    
    else if (iEndPos > iStrLen - 1) 
        return -1;  
        
    // If less than to lower boundary return not found   
    else if (iEndPos < 0) 
        return - 1;      
 
    sTmpStr = sStr[0 .. iEndPos + 1];

    return std.string.rfind(sTmpStr, sSubStr);  
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.rfindx( in char[], in char[], in int = -1 ).unittest"c );

    char[] sPlts = "Mars: the fourth Rock (Planet) from the Sun."c;
    int    i;

    i = rfindx("Abba"c,"a"c, 5);
    assert( i == -1 );
    i = rfindx(null, "a"c);
    assert( i == -1 );
    i = rfindx("def"c, "a"c);
    assert( i == -1 );
    i = rfindx("def"c, "f"c);
    assert( i == 2 );
    i = rfindx("dfefFfg"c, "fFf"c, 6);
    assert( i == 3 );
    i = rfindx("dfeffgffF"c, "ffF"c);
    assert( i == 6 );

    i = rfindx(sPlts, "from"c, 39);
    assert( i == 31 );
    i = rfindx(sPlts, "M"c, 15);
    assert( i == 0 );
    i = rfindx(sPlts, "Su"c);
    assert( i == 40 );
 }

/+
 ' char[] ireplace( in char[], in char[], in char[] )
 '
 ' Replace occurrences of sFrom[] with sTo[] in s[].
 '
 ' Note: Is Non-Case Sensitive.
 '
 +/
char[] ireplace(in char[] s, in char[] sFrom, in char[] sTo)
{
    char   c;
    char[] sReturn;
    uint   uiSLen    = s.length;
    uint   uiFromLen = sFrom.length;
    uint   uiToLen   = sTo.length;
    int    ip;
    int    ix;
    
    // if the length of any of the strings equals zero return "s",
    // except the "to" can be an "" (Empty String).
    if (uiFromLen == 0 || uiSLen == 0 || 
        uiFromLen > uiSLen - 1 || uiToLen > uiSLen - 1) 
        return s; 
       
    // find the position of the string from's 1st wchar
    c  = sFrom[0];    
    ip = std.string.ifind(s, c);
        
    // if the string from's 1st char was not found, return "s" 
    if (ip == -1 || ip + uiFromLen > uiSLen - 1) 
        return s;
    
    if (ip > 0) 
        sReturn = s[0 .. ip]; 
    
    for (ix = ip; ix < (uiSLen - uiFromLen) + 1; ix++)
    {
        if (std.string.icmp(s[ix .. (ix + uiFromLen)], sFrom) == 0) 
        {
            sReturn ~= sTo;
            ix += uiFromLen - 1;
        }
        else
            sReturn ~= s[ix .. (ix + 1)];
    }
        
    if (ix < uiSLen - 1) 
    {
        ip = uiSLen - ix;
        sReturn ~= s[ix .. ix + ip];     
    }

    return sReturn; 
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.ireplace( in char[], in char[], in char[] ).unittest"c );

    char[] s     = "This is a FOO foO list"c;
    char[] sFrom = "fOo"c;
    char[] sTo   = "silly"c;
    char[] sR;
    int    i;

    sR = ireplace(s, sFrom, sTo);
    i  = std.string.icmp(sR, "This is a silly silly list"c);
    assert( i == 0 );
    
    s = "Love is a two way Street!"c;
    sR = ireplace(s, "TwO"c, "one"c);
    assert( sR == "Love is a one way Street!"c );
    
    sR = ireplace("ApPLe-Bee"c, "pple"c, "te-A"c);
    assert( sR == "Ate-A-Bee"c );
}

/+
 ' int icount( in char[], in char[] )
 '
 ' Count up all instances of sSub[] in s[].
 '
 ' Note: Is Non-Case Sensitive.
 '
 +/
int icount(in char[] s, in char[] sSub)
{
    uint uiSLen   = s.length;
    uint uiSubLen = sSub.length;
    int  ip       = 0;
    int  iCounter = 0;

    if (uiSLen == 0 || uiSubLen == 0 || uiSubLen > uiSLen) 
        return 0;
    
    for (int ix = 0; ix < uiSLen; ix += ip + uiSubLen)
    {
        ip = std.string.ifind(s[ix .. uiSLen], sSub);

        if (ip == -1) 
            break;
    
        iCounter++;
    }
    
    return iCounter;
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.icount( in char[], in char[] ).unittest"c );

    char[] s    = "This iS a foFofof liSt"c;
    char[] sSub = "FOf"c;
    int    i;

    i = icount(s, sSub);
    assert( i == 2 );

    i = icount(s, "S"c);
    assert( i == 3 );
}

/+
 ' bool isASCIIString( in char[], in bool = false )
 '
 ' returns true if an ASCII character is found.
 '
 +/
bool isASCIIString(in char[] s, in bool bIgnoreWhiteSpace = false)
{
    /+
     ' Returns true if any character below 0x7F is found...    
     ' these are the characters that fall into the ASCII 
     ' 7-bit range of characters. Unless the Ignore White 
     ' Space is set to true, then bypass checking those characters.
     +/
    for (int ix = 0; ix < s.length; ix++)
    {
        if (bIgnoreWhiteSpace == true)
            if (std.string.find(WHITESPACE, s[ix]) != -1)
                continue;
                
        if (s[ix] < 0x0080) 
            return true;
    }
    
    return false;
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.isASCIIString( in char[], in bool = false ).unittest"c );

    char[] s1 = "\u0392\u1ECE\u0154\u0419 "c;
    char[] s2;
    bool   bFlag = false;
    
    bFlag = isASCIIString(s1, true);
    assert( bFlag == false );
    bFlag = isASCIIString(s1);
    assert( bFlag == true );
    bFlag = isASCIIString("Apple-\u03B2"c ~ "ee"c);
    assert( bFlag == true );
}

/+
 ' Function      : bool isRomanNumeral( in char )
 ' Date Created  : 03.Jun.04 
 ' Date Modified : 06.Sep.04 - Tested and complied with dmd v0.101
 ' Requirements  : none
 '
 +/
bool isRomanNumeral(in char cChar)
{
    return (std.string.find(ROMAN, cChar) != -1) ? true : false;
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.isRomanNumeral( in char ).unittest"c );

    bool bFlag = false;
    
    bFlag = isRomanNumeral('x');
    assert( bFlag == true );
    bFlag = isRomanNumeral('M');
    assert( bFlag == true );
    bFlag = isRomanNumeral('a');
    assert( bFlag == false );
}

/+
 ' Function      : bool isRomanNumeral( in char[] )
 ' Date Created  : 03.Jun.04 
 ' Date Modified : 06.Sep.04 - Tested and complied with dmd v0.101
 ' Requirements  : none
 '
 +/
bool isRomanNumeral(in char[] sStr)
{
    foreach(int iStrPos, char cChar; sStr)
        if (std.string.find(ROMAN, cChar) == -1) 
            return false;
      
    return true;
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.isRomanNumeral( in char[] ).unittest"c );

    char[] s1 = "MMVII"c;
    bool   bFlag = false;
    
    bFlag = isRomanNumeral(s1);
    assert( bFlag == true );
    bFlag = isRomanNumeral("XiIi"c);
    assert( bFlag == true );
    bFlag = isRomanNumeral("XiXa"c);
    assert( bFlag == false );
}

/+
 ' Function      : char[] toSoundex( in char[] )
 ' Version       : v0.1 Beta
 ' Author        : David L. 'SpottedTiger' Davis
 ' Date Created  : 14.Oct.04 Compiled and Tested with dmd v0.102
 ' Date Modified : (None)
 '               :
 ' Licence       : Public Domain
 ' --------------------------------------------------------------
 ' Soundex Rules based on this Web Site:
 ' http://www.archives.gov/research_room/genealogy/census/soundex.html
 ' --------------------------------------------------------------
 '
 ' Return the soundex code for a given 
 ' string based upon its phonetic sound.
 +/
char[] toSoundex(in char[] s) 
{
    char[] sWord = std.string.toupper(std.string.strip(s));
    char[] sSoundex;
    char   cPrevVal;
    char   cPrevVal2;

    /+ Return a transformed (A thru Z) ASCII alphabetic  
     ' character to its numeric appropriate Soundex Code 
     ' value or special character.
     '                 A,E,I,O,U,Y     = '/'
     '                 H,W             = '*'
     '                 B,F,P,V         = '1'
     '                 C,G,J,K,Q,S,X,Z = '2'
     '                 D,T             = '3'
     '                 L               = '4'
     '                 M,N             = '5'
     '                 R               = '6'
     '                 All others      = '-'
     +/
    static char getSoundexFormattingCode(in char c) 
    {
        switch (c) 
        {
          case 'A','E','I','O','U','Y': 
              return '/';  
              
          case 'H','W': 
              return '*'; 
              
          case 'B','F','P','V': 
              return '1';
              
          case 'C','G','J','K','Q','S','X','Z':
              return '2';
              
          case 'D','T':
              return '3';
        
          case 'L':
              return '4';
            
          case 'M','N':
              return '5';
            
          case 'R':
              return '6';
            
          default:    
              break;
        }

        return '-'; 
        
    } // end char getSoundexFormatingCode( in char )    
    
    // Keep the first alphabetic character
    sSoundex.length = 1;
    sSoundex[0] = sWord[0];
    
    // Transform the characters to a Soundex Code or 
    // a special processing character.
    for (uint uix = 0; uix < sWord.length; uix++)
        sWord[uix] = getSoundexFormattingCode(sWord[uix]);
    
    // Apply a number of the Soundex rules for double characters
    // and side-to-side Soundex code issues
    for (uint uix = 1; uix < sWord.length; uix++)
    {
        //writef( "%s", sWord[ uix ] );

        if (sWord[uix - 1] == sWord[uix] || cPrevVal == sWord[uix])
        {
           cPrevVal     = sWord[uix];
           sWord[ uix ] = '-';
        }    
        
        if (uix - 2 >= 0)
            if (cPrevVal2 == '*')
                if (sWord[uix - 2] == sWord[uix])
                    sWord[uix] = '-';

        if (sWord[uix] == '/' && cPrevVal != '*')
            cPrevVal = '\x20';

        cPrevVal2 = sWord[uix];
    }

    // Filter out the unwanted character to create the final Soundex String
    for (uint uix = 1; uix < sWord.length; uix++)
        if (sWord[uix] != '/' && sWord[uix] != '*' && sWord[uix] != '-')
            sSoundex ~= sWord[uix];            

    // If Soundex Code is shorter then four characters append four zeros.
    if (sSoundex.length < 4)
        sSoundex ~= "0000"c;
    
    //writefln();
    
    // Return only the first four characters
    return sSoundex[0 .. 4]; //.dup; 
}

unittest
{
    debug( stringplus ) writefln( "stringplus.toSoundex( in char[] ).unittest"c );

    // Sound-Alike Names 
    assert( "F626"c == toSoundex("fraser"c) );
    assert( "F626"c == toSoundex("frazier"c) );
    
    assert( "V230"c == toSoundex("Vogt"c) );
    assert( "V230"c == toSoundex("Voght"c) );
    assert( "V230"c == toSoundex("Voight"c) );
    
    assert( "F230"c == toSoundex("Fogt"c) );
    assert( "F230"c == toSoundex("Foght"c) );
    assert( "F230"c == toSoundex("Foight"c) );
    //----------------------------------------
    
    // Normal examples
    assert( "W252"c == toSoundex("Washington"c)  );
    assert( "E460"c == toSoundex("Euler"c) );
    assert( "H416"c == toSoundex("Hilbert"c) );
    assert( "K530"c == toSoundex("Knuth"c) );
    assert( "L222"c == toSoundex("Lukasiewicz"c) );
    assert( "V532"c == toSoundex("VanDeusen"c) );
    assert( "N253"c == toSoundex("Naesmyth"c) );  
    assert( "B630"c == toSoundex("Baird"c) );

    // When 2 or more of the same key letters (or their number 
    // equivalents) come together, they should be treated as 
    // ONE key letter.
    assert( "J250"c == toSoundex("Jackson"c) );
    assert( "S163"c == toSoundex("Sheppard"c) );
    assert( "S200"c == toSoundex("Sacks"c) );
    assert( "S640"c == toSoundex("Sherrell"c) );
    assert( "L300"c == toSoundex("Lloyd"c) );
    assert( "G362"c == toSoundex("Gutierrez"c) ); 
    assert( "C450"c == toSoundex("Callahan"c) );
    assert( "G200"c == toSoundex("Gauss"c) );

    // A key letter or its equivalent immediately following an 
    // initial letter of the same value should be disregarded 
    // in coding.
    assert( "P236"c == toSoundex("Pfister"c) );
    assert( "S232"c == toSoundex("Schechwitz"c) );
    assert( "S240"c == toSoundex("Szakal"c) );
    assert( "S460"c == toSoundex("Scklar"c) ); 
    
    // Key letters or their equivalents separated by H or W 
    // are coded as one key letter.   
    assert( "S200"c == toSoundex("Sokwzy"c) ); 
    assert( "S452"c == toSoundex("Schkolink"c) ); 
    assert( "S460"c == toSoundex("Schklar"c) ); 
    assert( "A261"c == toSoundex("Ashcraft"c) );
    assert( "S432"c == toSoundex("Schultz"c) );
    
    // When a repeated key letter or its equivalent is separated 
    // by an A, E, I, O, U or Y, the key letters or their 
    // equivalents are considered separately.
    assert( "T522"c == toSoundex("Tymczak"c) );
    assert( "S335"c == toSoundex("Staten"c) ); 
    assert( "S422"c == toSoundex("Salkiewicz"c) ); 
    assert( "S550"c == toSoundex("Simone"c) ); 
    
    // Some names do not contain any of the six key letters or their 
    // equivalents - that is they have only vowels A, E, I, O, U or 
    // W, H, Y, after the first letter of the name. Such names are 
    // filed in a group by themselves in back of their letter guide.
    assert( "S000"c == toSoundex("Shea"c) ); 
    assert( "L000"c == toSoundex("Lee"c) );
    
    // Testing variations of the Womack surname 
    assert( "W520"c == toSoundex("Wamac"c) ); 
    assert( "W520"c == toSoundex("Wamack"c) ); 
    assert( "W520"c == toSoundex("Wammack"c) ); 
    assert( "W520"c == toSoundex("Wammuck"c) ); 
    assert( "W520"c == toSoundex("Womac"c) ); 
    assert( "W520"c == toSoundex("Womach"c) ); 
    assert( "W520"c == toSoundex("Womich"c) ); 
    assert( "W520"c == toSoundex("Womick"c) ); 
    assert( "W520"c == toSoundex("Wommac"c) ); 
    assert( "W520"c == toSoundex("Wommack"c) );    
    assert( "W520"c == toSoundex("Wommuck"c) );    
    assert( "W520"c == toSoundex("Wonack"c) );    
    assert( "W520"c == toSoundex("Woomack"c) );     
}

/+
 ' Function      : char[] addOrdSuffix( in ulong )
 ' Author        : David L. 'SpottedTiger' Davis
 ' Date Created  : 25.May.02 in VB6
 ' Date Modified : 15.Nov.04 rewritten in D
 ' Licence       : Pubic Domain
 ' -----------------------------------------------
 ' Returns the passed in number as a string with a 
 ' ordinal suffix (st, nd, rd, th). 
 +/
char[] addOrdSuffix(in ulong ulNumber)
{
    char[] sNumber = std.string.toString(ulNumber);
    
    // Handles these numbers as special cases.
    if (ulNumber == 11 || ulNumber == 12 || ulNumber == 13)
        return sNumber ~ "th"c;

    switch (sNumber[sNumber.length - 1 .. sNumber.length])
    {
      case "1"c: return sNumber ~ "st"c;
      case "2"c: return sNumber ~ "nd"c;
      case "3"c: return sNumber ~ "rd"c;
      default  : return sNumber ~ "th"c;
    }
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.addOrdSuffix( in ulong ).unittest"c );
    assert( addOrdSuffix(123) == "123rd"c );
    assert( addOrdSuffix(21) == "21st"c );
    assert( addOrdSuffix(11) == "11th"c );
    assert( addOrdSuffix(12) == "12th"c );
    assert( addOrdSuffix(13) == "13th"c );
    assert( addOrdSuffix(1022) == "1022nd"c );
    assert( addOrdSuffix(1024) == "1024th"c );

    assert( addOrdSuffix(long.max) == "9223372036854775807th"c );
    assert( addOrdSuffix(ulong.min) == "0th"c );
    assert( addOrdSuffix(ulong.max) == "18446744073709551615th"c );
}

/+
 ' Function     : bool isDigits( in char[] )
 '              :           
 '              : Checks a string to see if all it's characters 
 '              : are digits, and if so it returns true.
 '              :
 ' Author       : David L. 'SpottedTiger' Davis
 ' Date Created : 19.Jun.04
 ' Language     : DigitalMars D v0.92
 '              :
 ' Requirements : std.string
 +/
bool isDigits(in char[] s)
{
    foreach (dchar c; s)
        if (std.string.find(digits, cast(char)c) == -1) 
            return false;
        
    return true;   
} 

unittest
{
    debug( stringplus ) writefln( "stringplus.isDigits( in char[] ).unittest"c );

    assert( isDigits("123"c) == true );
    assert( isDigits("0x59y"c) == false ); 
}

// Function : xtoul() - Hex string to an unsigned decimal whole number 
// Author   : David 'SpottedTiger' L. Davis
// Date     : 03.May.05
// CPU      : Intel x86 (Numercials stored as "Low byte to High byte")
// ----------------------------------------
// HexToDec() 
// Converts a Hex string from 0x0 to 0xFFFFFFFFFFFFFF
// into a ulong value 0 to 18,446,744,073,709,551,615
// also it handles the lowercase 'a' thru 'f'.
ulong xtoul(in char[] sx)
{
    ulong ul = 0L;
    int   j = 7;
    int   len = sx.length;
    char  c, c1, c2;
    
    const char[] zeros = "0000000000000000"c;
    union u { ulong ul; char[8] c; }
    
    u U;

    if (len == 0 || len > 16)
        throw new Exception( "xtoul() the string parameter is either an empty string,"c ~
                             " or its length is greater than 16 characters."c );
     
    // isHex()                         
    for (int i = 0; i < sx.length; i++)                             
    {
        //c = ( sx[i] > 'F' ? sx[i] - 32 : sx[i] );
        c = sx[i];

        if ((c >= '0' && c <= '9') || 
            (c >= 'A' && c <= 'F') ||
            (c >= 'a' && c <= 'f'))
           continue;
        else     
          throw new Exception("xtoul() invalid hex character used."c);   
    }
          
    if (len < 16)
        sx = zeros[0..16 - len] ~ sx;
    
    j = 7;
    for (int i = 0; i < 16; i += 2)
    {
        c1 = (sx[i] > 'F' ? sx[i] - 32 : sx[i]); 
        c2 = (sx[i + 1] > 'F' ? sx[i + 1] - 32 : sx[i + 1]);
        c1 = cast(int)(c1 > 52 ? c1 - 55 : c1 - 48) << 4;
        U.c[j--] = c1 + cast(int)(c2 > 52 ? c2 - 55 : c2 - 48);
    }

    return U.ul;
} 


unittest
{
    debug( stringplus ) writefln( "stringplus.xtoul( in char[] ).unittest"c );
    ulong ul;

    ul = xtoul("0"c); 
    assert( ul == 0x0 );
    ul = xtoul("FF"c); 
    assert( ul == 0xFF );
    ul = xtoul("eea"c); 
    assert( ul == 0xEEA );
    ul = xtoul("AB"c);
    assert( ul == 0xAB );
    ul = xtoul("ABCD"c);
    assert( ul == 0xABCD );
    ul = xtoul("A12CD00"c);
    assert( ul == 0xA12CD00 );
    ul = xtoul("FFFFFFFFFFFFFFFF"c); 
    assert( ul == 0xFFFFFFFFFFFFFFFF );
}

// Function : ultox() - Decimal unsigned whole number to Hex string
// Author   : David 'SpottedTiger' L. Davis
// Date     : 04.May.05
// CPU      : Intel x86 (Numercials stored as "Low byte to High byte")
// -----------------------------------------
// DecToHex()
// Accepts any positive number from 0 to 18,446,744,073,709,551,615
// and the returns an even number of hex strings up to 16 characters
// (from 0x0 to 0xFFFFFFFFFFFFFF).
char[] ultox(in ulong ul)
{
    char[16] sx;
    char     c1, c2;
    union    u { ulong ul; char[8] c; }
    int      i = 0, j = 0, k = 0;
    bit      z = true;
    u U;

    U.ul = ul;
    
    for (i = 7; i >= 0; i--)
    {
        c1 = U.c[i] >> 4;
        c1 = cast(char)(c1 > 9 ? c1 + 55 : c1 + 48);
        c2 = U.c[i] & 0x0F;
        c2 = cast(char)(c2 > 9 ? c2 + 55 : c2 + 48);
        
        if (z && c1 == '0' && c2 == '0')
            continue;

        z = false;
        sx[j++] = c1;
        sx[j++] = c2;
    }
    
    if (j > 0)
        //Copying a fixed array into a dynamic array, must COW.
        return sx[0..j].dup;
    else
        return "00"c;    
}

unittest
{
    debug( stringplus ) writefln( "stringplus.ultox( in ulong ).unittest"c );
    char[] sx;

    sx = ultox(0); //0x0
    assert( sx == "00"c );

    sx = ultox(255); //0xFF
    assert( sx == "FF"c );
    sx = ultox(171); //0xAB
    assert( sx == "AB"c );
    sx = ultox(43981); //0xABCD
    assert( sx == "ABCD"c );
    sx = ultox(169004288); //0xA12CD00
    assert( sx == "0A12CD00"c );
    sx = ultox(ulong.max); //0xFFFFFFFFFFFFFFFF
    assert( sx == "FFFFFFFFFFFFFFFF"c );
}

bool isHex(in char[] sx)   
{          
    char c;
                
    for (int i = 0; i < sx.length; i++)                             
    {
        c = sx[i];
        
        if ((c >= '0' && c <= '9') || 
            (c >= 'A' && c <= 'F') ||
            (c >= 'a' && c <= 'f'))
            continue;
        else     
            return false;   
    }
    
    return true;
}

unittest
{
    debug( stringplus ) writefln( "stringplus.isHex( in char[] ).unittest"c );
    
    assert( isHex("00"c) );
    assert( isHex("FF"c) );
    assert( isHex("Ffae0"c) );
    assert( isHex("AB"c) );
    assert( isHex("abdef"c) );
    assert( isHex("ABCD"c) );
    assert( isHex("0A12CD00"c) );
    assert( isHex("FFFFFFFFFFFFFFFF"c) );
    
    assert( isHex("00ER"c) != true );
    assert( isHex("0xW"c) != true );
}

debug( stringplus )
{
    int main()
    {
        writefln( "unittest done"c );
        return 0;    
    }
}